@@ -1,26 +1,81 @@
+# region common packages import and settings
+from pathlib import Path
+import uuid
+import json
+import os
+# endregion
+
+# region prepare config
+config_file_path = Path.home() / '.azdsdr_conf.json'
+
+# make sure the existence of a configure
+if not Path.exists(config_file_path):
+    with open(config_file_path,'w') as f:
+        f.write("{}")
+
+with open(config_file_path,'r') as f:
+    config_obj = json.load(f)
+
+def update_config(key,value):
+    config_obj[key] = value
+    config_obj_json = json.dumps(config_obj)
+    with open(config_file_path,'w') as f:
+        f.write(config_obj_json)
+# endregion
+
 # region Dremio
 import pandas as pd
 import warnings
 
 class DremioReader:
-    def __init__(self,username,token,host = "dremio-mcds.trafficmanager.net") -> None:
+    def __init__(
+        self
+        ,username
+        ,token      = None
+        ,host       = "dremio-mcds.trafficmanager.net"
+        ,port       = 31010
+        ,driver     = "Dremio Connector"
+    ) -> None:
         import pyodbc
         '''
         Initialize the Dremio connection, the connection object will be saved for sql queries
 
         Args:
             username (str): your dremio login email in format <alias>@microsoft.com
-            token (str): steps to generate the token: click username -> Account Settings -> Personal Access Tokens
+            token (str): steps to generate the token: click username -> Account Settings -> Personal Access Tokens. 
+                         The token need to be provided at the first time using it. the token will be cached in file 
+                         `~/.azdsdr_conf.json` file.
+            host (str): your target dremio host
+            port (int): the port to connect dremko, default value set as 31010
+            driver (str): the odbc driver name you give when you setup the Dremio odbc driver. 
+        
+        Example: 
+            ```
+            from azdsdr.readers import DremioReader
+            username    = "abc@abc.com"
+            dr          = DremioReader(username=username)
+            ```
         '''
-        port    = 31010
-        uid     = username
-        token   = token
-        driver  = "Dremio Connector"
-
-        self.connection = pyodbc.connect(
-            f"Driver={driver};ConnectionType=Direct;HOST={host};PORT={port};AuthenticationType=Plain;UID={uid};PWD={token};ssl=1"
-            ,autocommit=True
-        )
+        # load token from configuration file if token is not provided. 
+        if not token:
+            token = config_obj['dremio_token']
+        else: 
+            # the token is provided in the parameter. add or update the original token
+            update_config('dremio_token',token)
+        
+        # raise exception if no token is provided. 
+        if not token:
+            raise Exception('No dremio token is found from config file either parameter.')
+        
+        # if no token is provided
+        try:
+            self.connection = pyodbc.connect(
+                f"Driver={driver};ConnectionType=Direct;HOST={host};PORT={port};AuthenticationType=Plain;UID={username};PWD={token};ssl=1"
+                ,autocommit=True
+            )
+        except:
+            self.connection = None
+            raise Exception('Connect to dremio via pyodbc error. Please check host, port, username and Dremio token.')
 
     def run_sql(self,sql_query:str) -> pd.DataFrame:
         '''
@@ -61,6 +116,7 @@ class KustoReader:
                 ,cluster            = "https://help.kusto.windows.net"
                 ,db                 = "Samples"
                 ,ingest_cluster_str = None
+                ,timeout_hours      = 1
                 ) -> None:
         '''
         Initilize Kusto connection with additional timeout settings
@@ -70,20 +126,21 @@ class KustoReader:
         self.kusto_client   = KustoClient(kcsb)
         self.properties     = ClientRequestProperties()
         self.properties.set_option(self.properties.results_defer_partial_query_failures_option_name, True)
-        self.properties.set_option(self.properties.request_timeout_option_name, timedelta(seconds=60 * 60 * 2))
+        self.properties.set_option(self.properties.request_timeout_option_name, timedelta(seconds=60 * 60 * timeout_hours))
         if ingest_cluster_str:
             self.ingest_cluster  = KustoConnectionStringBuilder.with_az_cli_authentication(ingest_cluster_str)
             self.ingest_client   = QueuedIngestClient(self.ingest_cluster)
     
     def run_kql(self,kql:str) -> pd.DataFrame:
         '''
-        Run the input Kusto script on target cluster and database
+        Run the input Kusto script on target cluster and database, This function
+        will return the first result set of execution. 
 
         Args:
             kql (str): the Kusto script in plain string
         
         Returns:
-            pd.Dataframe: pandas Dataframe containing results of SQL query from Dremio
+            pd.Dataframe: pandas Dataframe containing results of Kusto.
         '''
         r_df = None
         try:
@@ -96,6 +153,30 @@ class KustoReader:
             traceback.print_exc()
             return None
         return r_df
+
+    def run_kql_all(self,kql:str) -> list:
+        '''
+        Run the input Kusto script on target cluster and database, This function
+        will return all result set
+
+        Args:
+            kql (str): the Kusto script in plain string
+        
+        Returns:
+            list: list of pd.Dataframe.
+        '''
+        r_df_list = []
+        try:
+            r_set = self.kusto_client.execute(database = self.db,query=kql,properties=self.properties).primary_results
+            for r in r_set:
+                r_df_list.append(dataframe_from_result_table(r))
+        except KustoServiceError as error:
+            print('something wrong')
+            print("Is semantic error:", error.is_semantic_error())
+            print("Has partial results:", error.has_partial_results())
+            traceback.print_exc()
+            return None
+        return r_df_list
     
     def is_table_exist(self,table_name)->bool:
         '''
@@ -389,18 +470,29 @@ from azure.storage.blob import (
     ,BlobBlock
 )
 from datetime import datetime,timedelta
-import uuid
 
 class AzureBlobReader:
     '''
+    Args:
+        * container_name is required 
+        * The class will retrieve blob_conn_str from configuration file is the parameter is set as None
+
+    Functions from this class support the following Features: 
     * Download file
     * Upload file
     * Get Blob SAS token
     * Get Blob SAS Url
     * Delete file
     '''
-    def __init__(self,blob_conn_str,container_name) -> None:
-        self.connect_string         = blob_conn_str
+    def __init__(self,container_name,blob_conn_str=None):
+        if blob_conn_str:
+            self.connect_string         = blob_conn_str
+            # update the azure blob connection string with the newest one
+            update_config('azure_blob_connstr',blob_conn_str)
+        else:
+            # get connect_string from configure file
+            self.connect_string = config_obj['azure_blob_connstr']
+
         self.blob_service_client    = BlobServiceClient.from_connection_string(self.connect_string)
         #self.blob_service_client.max_single_put_size = 4*1024*1024              # 4M
         #self.blob_service_client.timeout = 60*20                                # 10 mins
@@ -414,7 +506,39 @@ class AzureBlobReader:
         with open(local_file_path,'wb') as f:
             download_stream = blob_client.download_blob()
             f.write(download_stream.readall())
-        return "download done"
+        return f"blob file {blob_file_path} is downloaded to {local_file_path}"
+
+    def download_file_list(self,blob_file_path_list,local_file_path) -> str:
+        '''
+        Download a list of file with the same schema
+        '''
+        # write the complete csv file with header
+        blob_client = self.container_client.get_blob_client(blob_file_path_list[0])
+        with open(local_file_path,'wb') as f:
+            download_stream = blob_client.download_blob()
+            f.write(download_stream.readall())
+
+        # write the following files without header
+        try:
+            for blob_file_path in blob_file_path_list[1:]:
+                blob_client = self.container_client.get_blob_client(blob_file_path)
+                # write to a temp file 
+                with open("temp.csv",'ab') as f:
+                    download_stream = blob_client.download_blob()
+                    f.write(download_stream.readall())
+                # remove teh first line and then write back to the target file 
+                with open(local_file_path,'a') as f_target:
+                    with open('temp.csv','r+') as f:
+                        lines = f.readlines()
+                        f.seek(0)
+                        f.truncate()
+                        f_target.writelines(lines[1:])
+        except:
+            raise Exception('write csv file error')
+        finally:
+            os.remove('temp.csv')
+
+        return f"All blob files are downloaded to {local_file_path}"
 
     def upload_file(self,blob_file_path,local_file_path):
         '''
@@ -486,14 +610,17 @@ class AzureBlobReader:
             blob_client.delete_blob()
         except:
             print('Delete file error')
+    
+    def delete_blob_files(self,blob_file_path_list):
+        try:
+            for f_path in blob_file_path_list:
+                self.delete_blob_file(f_path)
+        except:
+            print('Delete file error')
 
 # endregion
 
 # region pipelines 
-
-from pathlib import Path
-import uuid
-
 class Pipelines:
     def __init__(self,**kwargs) -> None:
         '''
@@ -506,12 +633,12 @@ class Pipelines:
             ,azure_blob_container
         '''
         if kwargs:
-            self.kusto_cluster         = kwargs['kusto_cluster']
-            self.kusto_cluster_ingest  = kwargs['kusto_cluster_ingest']                   
-            self.kusto_db              = kwargs['kusto_db']   
-            self.dremio_user_name      = kwargs['dremio_user_name']           
-            self.dremio_host           = kwargs['dremio_host']       
-            self.azure_blob_container  = kwargs['azure_blob_container']                            
+            self.kusto_cluster         = kwargs.get('kusto_cluster','')
+            self.kusto_cluster_ingest  = kwargs.get('kusto_cluster_ingest','')                   
+            self.kusto_db              = kwargs.get('kusto_db','')
+            self.dremio_user_name      = kwargs.get('dremio_user_name','')           
+            self.dremio_host           = kwargs.get('dremio_host','')       
+            self.azure_blob_container  = kwargs.get('azure_blob_container','')                            
 
     def load_azure_blob_context(self):
         '''
@@ -522,14 +649,7 @@ class Pipelines:
 
         If no `.azureblob_constr` exists. raise an error message. 
         '''
-        con_str_path    = Path.home() / '.azureblob_constr'
-
-        if not Path.exists(con_str_path):
-            print('No Azure Blob connection ')
-            sys.exit(0)
-
-        with open(con_str_path,'r') as f:
-            con_str = f.read()
+        con_str = config_obj['azure_blob_connstr']
 
         self.abr = AzureBlobReader(
             blob_conn_str   = con_str
@@ -541,14 +661,7 @@ class Pipelines:
         '''
         The function will load Dremio token from configure file .dremio_token.
         '''
-        con_str_path    = Path.home() / '.dremio_token'
-        
-        if not Path.exists(con_str_path):
-            print('No dremio token detected, please double check ~/.dremio_token file')
-            sys.exit(0)
-        
-        with open(con_str_path,'r') as f:
-            token = f.read()
+        token = config_obj['dremio_token']
         
         self.dr  = DremioReader(
             username    = self.dremio_user_name
@@ -704,5 +817,88 @@ class Pipelines:
             self.abr.delete_blob_file(csv_file_name)
 
         print('all done')
+    
+    def kusto_to_csv(
+        self
+        ,input_kql
+        ,output_csv_file_name
+    ):
+        '''
+        The function will 
+        1. execute plain KQL(without .export async to csv). 
+        2. output data to Azure blob storage as csv file.
+        3. download data from azure blob storage to local file path
 
+        Args:
+            input_kql (str): the kusto script that generate the dataset
+            output_csv_path (str): the local csv file path (absolute or relative path)
+        
+        Returns: 
+            str: the execution status of the function. 
+        
+        Example: 
+            [TODO]
+        '''
+        self.load_kusto_context()
+        self.load_azure_blob_context()
+        # load Azure blob key str from configure file
+        try:
+            azure_blob_key = config_obj['azure_blob_key']
+            azure_blob_account = config_obj['azure_blob_connstr'].split(';')[1].replace('AccountName=','')
+            print(f"azure_blob_account {azure_blob_account} will be used as the exporting middle layer")
+        except:
+            print('get azure blob key and connection string error')
+
+        # temp name for azure blob 
+        temp_name = str(uuid.uuid4())
+        
+        kusto_head = f'''.export async to csv(
+            h@"https://{azure_blob_account}.blob.core.windows.net:443//{self.azure_blob_container}/azdsdr;{azure_blob_key}"
+        ) with (
+            sizeLimit        = 100000000
+            ,namePrefix      = {temp_name}
+            ,includeHeaders  = all
+            ,encoding        = UTF8NoBOM
+            ,distributed     = false
+        )
+        <|
+        '''
+
+        try:
+            kql = f"""{kusto_head}{input_kql}"""
+            print(kql)
+            # how to detect kusto run status? 
+            r = self.kr.run_kql(kql)
+            op_id = r['OperationId'][0]
+            print(op_id)
+
+            # use a loop to check the status of the kusto execution
+            check_kql = f".show operations {op_id}"
+            while(True):
+                r = self.kr.run_kql(check_kql)
+                time.sleep(1)
+                state = r['State'][0]
+                if state == "Completed":
+                    print('Kusto export done')
+                    break
+            
+            print('Kusto export to Azure Blob done.')
+        except:
+            raise Exception('Kusto export error')
+
+        try:
+            # query the export files start with temp uuid. 
+            file_list = self.abr.container_client.list_blobs(name_starts_with = f"azdsdr/{temp_name}")
+            file_name_list = [f['name'] for f in file_list]
+            self.abr.download_file_list(
+                blob_file_path_list = file_name_list
+                ,local_file_path    = output_csv_file_name
+            )
+        except:
+            raise Exception('blob csv files download error')
+        finally:
+            # delete blob temp file
+            self.abr.delete_blob_files(blob_file_path_list=file_name_list)
+
+        print('Kusto to CSV done!')
 # endregion
\ No newline at end of file
